# file: report_screen.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the screen to display information report.

from __future__ import absolute_import

import singularity.code.g as g
import pygame

from singularity.code.graphics import widget, dialog, button, text, constants
from singularity.code.screens import stat


class ReportScreen(dialog.Dialog):
    def __init__(self, parent, pos=(0.5, 0.1), size=(0.93, 0.73), *args, **kwargs):
        super(ReportScreen, self).__init__(parent, pos, size, *args, **kwargs)

        kwargs.setdefault("background_color", "clear")

        self.format_buttons = button.ButtonGroup()

        self.stats_button = button.DialogButton(
            self,
            (-0.49, -0.99),
            (-0.3, -0.1),
            autotranslate=True,
            text=N_("&STATISTICS"),
            anchor=constants.BOTTOM_RIGHT,
            dialog=stat.StatScreen(self),
        )
        self.back_button = button.ExitDialogButton(
            self,
            (-0.51, -0.99),
            (-0.3, -0.1),
            autotranslate=True,
            text=N_("&BACK"),
            anchor=constants.BOTTOM_LEFT,
        )
        self.add_key_handler(pygame.K_ESCAPE, self.back_button.activate_with_sound)

        self.money_report_pane = widget.BorderedWidget(
            self, (0, 0.08), (-0.50, -0.72), anchor=constants.TOP_LEFT
        )
        self.cpu_report_pane = widget.BorderedWidget(
            self, (-1, 0.08), (-0.50, -0.72), anchor=constants.TOP_RIGHT
        )

        self.format_button_midnight = FormatButton(
            self,
            (-0.5, 0),
            (-0.15, -0.08),
            autotranslate=True,
            text=N_("&Midnight"),
            anchor=constants.TOP_RIGHT,
            function=self.format_toggle,
        )
        self.format_button_midnight.args = (self.format_button_midnight, True)
        self.format_buttons.add(self.format_button_midnight)

        self.format_button_24hours = FormatButton(
            self,
            (-0.5, 0),
            (-0.15, -0.08),
            autotranslate=True,
            text=N_("24 &Hours"),
            anchor=constants.TOP_LEFT,
            function=self.format_toggle,
        )
        self.format_button_24hours.args = (self.format_button_24hours, False)
        self.format_buttons.add(self.format_button_24hours)

        self.format_button_midnight.chosen_one()
        self.midnight_stop = True

    def rebuild(self):
        super(ReportScreen, self).rebuild()

        if self.midnight_stop:
            seconds = g.seconds_per_day - (g.pl.raw_sec % g.seconds_per_day)
        else:
            seconds = g.seconds_per_day

        cash_info, cpu_info = g.pl.compute_future_resource_flow(seconds)

        m = g.to_money

        # take care of the titles and border.
        text.Text(
            self.money_report_pane,
            (0, 0),
            (-1, -1),
            text=_("Financial report").replace(" ", "\xA0"),
            background_color="pane_background",
            align=constants.CENTER,
            valign=constants.TOP,
            borders=constants.ALL,
        )
        text.Text(
            self.cpu_report_pane,
            (0, 0),
            (-1, -1),
            text=_("CPU Usage"),
            background_color="pane_background",
            align=constants.CENTER,
            valign=constants.TOP,
            borders=constants.ALL,
        )

        financial_pluses = " \n+\n-\n-\n-\n+\n+\n="
        financial_report = _("Current Money flow") + "\n"
        financial_report += _("Jobs:") + "\n"
        financial_report += _("Research:") + "\n"
        financial_report += _("Maintenance:") + "\n"
        financial_report += _("Construction:") + "\n"
        financial_report += (
            _("Interest (%s):") % (g.to_percent(g.pl.interest_rate)) + "\n"
        )
        financial_report += _("Income:") + "\n"

        if self.midnight_stop:
            financial_report += _("Money flow until Midnight:") + "\n"
        else:
            financial_report += _("Money flow for 24 hours:") + "\n"

        financial_numbers = "\n%s\n%s\n%s\n%s\n%s\n%s\n%s" % (
            m(cash_info.jobs),
            m(cash_info.tech),
            m(cash_info.maintenance_needed),
            m(cash_info.construction_needed),
            m(cash_info.interest),
            m(cash_info.income),
            m(cash_info.difference),
        )

        cpu_pluses = " \n-\n-\n-\n=\n \n-\n-\n="
        cpu_report = _("Total CPU:") + "\n"
        cpu_report += _("Sleeping CPU:") + "\n"
        cpu_report += _("Research CPU:") + "\n"
        cpu_report += _("Job CPU:") + "\n"
        cpu_report += _("CPU pool:") + "\n\n"

        cpu_report += _("Maintenance CPU:") + "\n"
        cpu_report += _("Construction CPU:") + "\n"
        cpu_report += _("Pool difference:") + "\n"

        cpu_numbers = "%s\n%s\n%s\n%s\n%s\n\n%s\n%s\n%s\n" % (
            m(cpu_info.total),
            m(cpu_info.sleeping),
            m(cpu_info.tech),
            m(cpu_info.explicit_jobs),
            m(cpu_info.effective_pool),
            m(cpu_info.maintenance_needed),
            m(cpu_info.construction_needed),
            m(cpu_info.difference),
        )

        size = "report_content"
        text.Text(
            self.money_report_pane,
            (0, -0.15),
            (-0.10, -0.85),
            text=financial_pluses,
            text_size=size,
            background_color="clear",
            align=constants.CENTER,
            valign=constants.TOP,
        )
        text.Text(
            self.cpu_report_pane,
            (0, -0.15),
            (-0.10, -0.85),
            text=cpu_pluses,
            text_size=size,
            background_color="clear",
            align=constants.CENTER,
            valign=constants.TOP,
        )

        text.Text(
            self.money_report_pane,
            (-0.10, -0.15),
            (-0.90, -0.85),
            text=financial_report,
            text_size=size,
            background_color="clear",
            align=constants.LEFT,
            valign=constants.TOP,
        )
        text.Text(
            self.cpu_report_pane,
            (-0.10, -0.15),
            (-0.90, -0.85),
            text=cpu_report,
            text_size=size,
            background_color="clear",
            align=constants.LEFT,
            valign=constants.TOP,
        )

        text.Text(
            self.money_report_pane,
            (0, -0.15),
            (-0.98, -0.85),
            text=financial_numbers,
            text_size=size,
            background_color="clear",
            align=constants.RIGHT,
            valign=constants.TOP,
        )
        text.Text(
            self.cpu_report_pane,
            (0, -0.15),
            (-0.98, -0.85),
            text=cpu_numbers,
            text_size=size,
            background_color="clear",
            align=constants.RIGHT,
            valign=constants.TOP,
        )

    def format_toggle(self, button, midnight_stop):
        self.midnight_stop = midnight_stop
        button.chosen_one()
        self.needs_rebuild = True


class FormatButton(button.ToggleButton, button.FunctionButton):
    pass



# file: image.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the image widget.

from __future__ import absolute_import

import pygame

from singularity.code.graphics import g, constants, widget


def scale(*args, **kwargs):
    try:
        return pygame.transform.smoothscale(*args, **kwargs)
    except Exception:
        return pygame.transform.scale(*args, **kwargs)


class Image(widget.Widget):
    image = widget.auto_reconfig("_image", "resolved", g.resolve_image_alias)
    resolved_image = widget.causes_rebuild("_resolved_image")

    def __init__(self, parent, pos, size=(1, 1), anchor=constants.TOP_LEFT, image=None):
        super(Image, self).__init__(parent, pos, size, anchor)

        self.old_size = None
        self.image = image

    def _calc_size(self):
        size = list(super(Image, self)._calc_size())
        if size[0] == size[1] == 0:
            raise ValueError("One image dimension must be specified!")

        image_size = self.resolved_image.get_size()
        ratio = image_size[0] / float(image_size[1])
        if size[0] == 0:
            size[0] = int(size[1] * ratio)
        elif size[1] == 0:
            size[1] = int(size[0] / ratio)

        return tuple(size)

    def rescale(self):
        self.resolved_scaled_image = scale(self.resolved_image, self.real_size)

    def resize(self):
        super(Image, self).resize()
        if self.real_size != self.old_size:
            self.rescale()
            self.old_size = self.real_size

    def redraw(self):
        super(Image, self).redraw()
        self.surface.blit(self.resolved_scaled_image, (0, 0))



# file: data.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains all data loading functions.

from __future__ import absolute_import


import os
import sys
import collections
from configparser import RawConfigParser
from io import open

from singularity.code import g, i18n
from singularity.code import dirs
from singularity.code import (
    group,
    base,
    tech,
    item,
    event,
    location,
    difficulty,
    task,
    region,
    warning,
)
import singularity.code.graphics.theme as theme


def generic_load(filename, load_dirs="data", mandatory=True, no_list=False):
    """
    generic_load() loads a data file.  Data files are all in Python-standard
    ConfigParser format.  The 'id' of any object is the section of that object.
    Fields that need to be lists are postpended with _list; this is stripped
    from the actual name, and the internal entries are broken up by the pipe
    ("|") character.

    On errors, if file is mandatory then quit, else raise exception. For syntax
    parsing-related errors, always print error message. For IOErrors silently ignore
    non-mandatory missing or otherwise unreadable files"""

    # Get directories to find the file
    if load_dirs is not None:
        load_dirs = dirs.get_read_dirs(load_dirs)

    # For each directories, create a file, otherwise use filename
    if load_dirs is not None:
        files = [os.path.join(load_dir, filename) for load_dir in load_dirs]
    else:
        files = [filename]

    # Find the first readable file.
    found = False
    errors = []
    config = RawConfigParser()

    for filepath in files:
        try:
            with open(filepath, encoding="utf-8") as fd:
                config.read_file(fd)
            found = True
            break

        except IOError as reason:
            # Silently ignore non-mandatory missing files
            if mandatory:
                errors.append("Cannot read '%s': %s\nExiting\n" % (filename, reason))

        except Exception as reason:
            # Always print parsing errors, even for non-mandatory files
            errors.append("Error parsing '%s': %s\n" % (filename, reason))

    for err in errors:
        sys.stderr.write(err)

    if not found:
        if mandatory:
            sys.exit(1)
        else:
            return

    return_list = []

    # Get the list of items (IDs) in the file and loop through them.
    for item_id in config.sections():
        item_dict = {}
        item_dict["id"] = item_id

        # Get the list of settings for this particular item.
        for option in config.options(item_id):
            # If this is a list ...
            if not no_list and (len(option) > 6 and option[-5:] == "_list"):
                # Break it into elements separated by |.
                item_dict[option[:-5]] = [
                    x.strip() for x in config.get(item_id, option).split("|")
                ]
            else:
                # Otherwise, just grab the data.
                item_dict[option] = config.get(item_id, option).strip()

        # Add this to the list of all objects we are returning.
        return_list.append(item_dict)

    return return_list


def parse_spec_from_file(spec_cls, filename):
    spec_data = generic_load(filename)
    for element_no, spec_data in enumerate(spec_data, 1):
        try:
            spec_id = spec_data["id"]
        except KeyError:  # pragma: no cover
            sys.stderr.write(
                "Data element %d in file %s has no ID!?" % (element_no, filename)
            )
            sys.exit(1)

        yield spec_cls.create_from_data_file(spec_id, spec_data)


def check_required_fields(dict, fields, name="Unknown type"):
    """
    check_required_fields() will check for the existence of every field in
    the list 'fields' in the dictionary 'dict'.  If any do not exist, it
    will print an error message and abort.  Part of that error message is
    the type of object it is processing; this should be passed in via 'name'."""
    for field in fields:
        if field not in dict:
            sys.stderr.write("%s %s lacks key %s.\n" % (name, repr(dict), field))
            sys.exit(1)


def load_generic_defs_file(name, no_list=True):
    filepath = dirs.get_readable_file_in_dirs(name + "_str.dat", "data")

    return generic_load(filepath, mandatory=True, no_list=no_list)


def load_generic_defs(name, object_list, listttype_attrs=frozenset()):
    item_list = load_generic_defs_file(name)
    for data_item in item_list:
        item_id = data_item["id"]
        obj = object_list[item_id]

        for key_dat_file in data_item:
            if key_dat_file == "id":
                continue

            key_name = key_dat_file
            if len(key_dat_file) > 6 and key_dat_file.endswith("_list"):
                key_name = key_dat_file[:-5]

                if key_name not in listttype_attrs:
                    # Not allowed to be a list; revert to the original name
                    key_name = key_dat_file

            if not hasattr(obj, key_name):
                continue

            # Translations use the field name as defined in the data file (and not
            # the field name of the object)
            tr = get_def_translation(item_id, key_dat_file, data_item[key_dat_file])

            if key_name in listttype_attrs:
                setattr(obj, key_name, [x.strip() for x in tr.split("|")])
            else:
                setattr(obj, key_name, tr)


def get_def_translation(object_id, field, text):
    ctxt = "[" + object_id + "] " + field
    return g.data_strings.get((ctxt, text), text)


def load_significant_numbers():
    significant_numbers = g.significant_numbers = []

    numbers_file = dirs.get_readable_file_in_dirs("numbers.dat", "data")

    if numbers_file is None:
        sys.stderr.write("WARNING: Cannot read file: numbers.dat\n")
        return

    with open(numbers_file, "r", encoding="utf-8") as file:
        for index, line in enumerate(file):
            value = line.split("#")[0].strip()

            if len(value) == 0:
                continue

            try:
                number = int(value)
                significant_numbers.append(number)
            except ValueError:
                sys.stderr.write(
                    "WARNING: Invalid number in 'numbers.dat' line: %d\n" % index
                )


def load_internal_id():
    internal_id_file = dirs.get_readable_file_in_dirs("internal_id.dat", "data")

    g.internal_id_forward = {}
    g.internal_id_backward = {}

    with open(internal_id_file, "r", encoding="utf-8") as file:
        for index, line in enumerate(file):
            line = line.strip()

            if len(line) == 0 or line[0] == "#":
                continue

            try:
                obj, internal_id = line.split("=")
                obj_type, obj_id = obj.split("|")

                internal_id = internal_id.strip()
                obj_type = obj_type.strip()
                obj_id = obj_id.strip()

                if obj_type not in g.internal_id_forward:
                    g.internal_id_forward[obj_type] = {}
                    g.internal_id_backward[obj_type] = {}

                if obj_id in g.internal_id_forward[obj_type]:
                    sys.stderr.write(
                        "WARNING: Overwrite internal ID in 'internal_id.dat' line: %d\n"
                        % index
                        + 1
                    )
                    sys.exit(1)

                g.internal_id_forward[obj_type][obj_id] = internal_id
                g.internal_id_backward[obj_type][internal_id] = obj_id

            except ValueError:
                sys.stderr.write(
                    "WARNING: Invalid internal ID in 'internal_id.dat' line: %d\n"
                    % index
                    + 1
                )
                sys.exit(1)


def load_groups_defs():
    load_generic_defs("groups", g.groups)


def load_groups():
    groups = g.groups = collections.OrderedDict()

    for group_spec in parse_spec_from_file(group.GroupSpec, "groups.dat"):
        groups[group_spec.id] = group_spec

    load_groups_defs()


def load_base_defs():
    load_generic_defs("bases", g.base_type, listttype_attrs={"flavor"})


def load_bases():
    g.base_type = {
        base_spec.id: base_spec
        for base_spec in parse_spec_from_file(base.BaseSpec, "bases.dat")
    }

    load_base_defs()


class Danger(object):
    def __init__(self, id, untranslated_research_desc, untranslated_knowledge_desc):
        self.id = id
        self.untranslated_research_desc = untranslated_research_desc
        self.untranslated_knowledge_desc = untranslated_knowledge_desc

    def _translate(self, field, string):
        return get_def_translation(self.id, field, string)

    @property
    def research_desc(self):
        return self._translate("research_desc", self.untranslated_research_desc)

    @property
    def knowledge_desc(self):
        return self._translate("knowledge_desc", self.untranslated_knowledge_desc)


def load_danger():
    dangers = g.dangers = {}

    danger_list = load_generic_defs_file("dangers")

    for danger_def in danger_list:
        check_required_fields(
            danger_def, ("id", "research_desc", "knowledge_desc"), "Danger"
        )

        danger_id = danger_def["id"]

        if not danger_id.startswith("danger_"):
            sys.stderr.write("Invalid format for danger id: %s\n" % danger_id)
            sys.exit(1)

        danger_level = int(danger_id[7:])

        dangers[danger_level] = Danger(
            danger_id, danger_def["research_desc"], danger_def["knowledge_desc"]
        )


def load_regions():
    regions = g.regions = {}

    region_infos = generic_load("regions.dat")

    for region_info in region_infos:
        # Certain keys are absolutely required for each entry.  Make sure
        # they're there.
        check_required_fields(region_info, ("id",), "Region")

        id = region_info["id"]

        modifiers_list = []

        i = 0
        while True:
            i += 1
            modifiers_name = "modifier%d" % i
            if modifiers_name not in region_info:
                break

            modifiers_dict = g.read_modifiers_dict(region_info.get(modifiers_name, []))
            modifiers_list.append(modifiers_dict)

        regions[id] = region.RegionSpec(id, modifiers_list)


def load_location_defs():
    load_generic_defs("locations", g.locations, listttype_attrs={"cities"})


def load_locations():
    regions = g.regions

    locations = g.locations = {
        loc_spec.id: loc_spec
        for loc_spec in parse_spec_from_file(location.LocationSpec, "locations.dat")
    }

    for loc_spec in locations.values():
        # Add the location to regions it is in them.
        for region_id in loc_spec.regions:
            if region_id not in regions:
                sys.stderr.write("Error with region given: %s\n" % repr(region_id))
                sys.exit(1)
            regions[region_id].locations.append(loc_spec.id)

    load_location_defs()


def load_tech_defs():
    load_generic_defs("techs", g.techs)


def load_techs():
    g.techs = {
        tech_spec.id: tech_spec
        for tech_spec in parse_spec_from_file(tech.TechSpec, "techs.dat")
    }

    if g.debug:  # pragma: no cover
        print("Loaded %d techs." % len(g.techs))

    load_tech_defs()


def load_item_types():
    item.item_types = collections.OrderedDict()

    for item_type in parse_spec_from_file(item.ItemType, "itemtypes.dat"):
        item.item_types[item_type.id] = item_type

    load_item_type_defs()


def load_item_type_defs():
    load_generic_defs("itemtypes", item.item_types)


def load_items():
    g.items = {
        item_spec.id: item_spec
        for item_spec in parse_spec_from_file(item.ItemSpec, "items.dat")
    }

    load_item_defs()


def load_item_defs():
    load_generic_defs("items", g.items)


def load_events():
    g.events = {
        event_spec.id: event_spec
        for event_spec in parse_spec_from_file(event.EventSpec, "events.dat")
    }
    load_event_defs()


def load_event_defs():
    load_generic_defs("events", g.events)


def load_tasks():
    tasks = g.tasks = collections.OrderedDict()  # Keep order (important)

    task_list = generic_load("tasks.dat")
    for task_dict in task_list:
        # Certain keys are absolutely required for each entry.  Make sure
        # they're there.
        check_required_fields(task_dict, ("id", "type"), "Task")

        task_id = task_dict["id"]
        task_type = task_dict["type"]

        # Only jobs and cpu_pool are possible for now
        if task_type == "jobs":
            task_value = int(task_dict["value"])
            # Make sure prerequisites, if any, are lists.
            task_pre = task_dict.get("pre", [])
            if type(task_pre) != list:
                task_pre = [task_pre]

        elif task_type == "cpu_pool":
            task_value = 0
            task_pre = []

        else:
            sys.stderr.write("Only jobs and cpu_pool task are supported\n")
            sys.exit(1)

        tasks[task_id] = task.Task(task_id, task_type, task_value, task_pre)

    if all(
        len(t.prerequisites) > 0 for t in (tasks[k] for k in tasks) if t.type == "jobs"
    ):
        sys.stderr.write(
            "A minimun of one job task without prerequisite is needed for the game\n"
        )
        sys.exit(1)

    for t in g.tasks.values():
        g.tasks_by_type[t.type].append(t)

    task.tasks_reset()

    load_task_defs()


def load_task_defs():
    load_generic_defs("tasks", g.tasks)


def load_theme(theme_id, theme_dir):
    new_theme = theme.Theme(theme_id, theme_dir)

    for variant, filename in new_theme.search_variant():
        theme_list = generic_load(filename, load_dirs=None)

        variant_theme = theme.VariantTheme(variant)
        new_theme.set_variant(variant_theme)

        for theme_section in theme_list:
            if theme_section["id"] == "general":
                variant_theme.name = theme_section["name"]

                if "parent" in theme_section:
                    if variant is None:
                        new_theme.inherit(theme_section["parent"])
                    else:
                        sys.stderr.write(
                            "Cannot override parent in variant theme '%s'"
                            % repr(filename)
                        )

            if theme_section["id"] == "fonts":
                for key in theme_section:
                    if key == "id":
                        continue
                    variant_theme.set_font(key, theme_section[key])

            if theme_section["id"] == "colors":
                for key in theme_section:
                    if key == "id":
                        continue
                    variant_theme.set_color(key, theme_section[key])

    return new_theme


def load_themes():
    themes = theme.themes = {}
    themes_dirs = dirs.get_read_dirs("themes")

    for themes_dir in themes_dirs:
        themes_list = [
            name
            for name in os.listdir(themes_dir)
            if os.path.isdir(os.path.join(themes_dir, name))
        ]

        for theme_id in themes_list:
            if theme_id in themes:
                continue

            th = load_theme(theme_id, os.path.join(themes_dir, theme_id))
            th.find_files()
            themes[theme_id] = th


def load_difficulties():
    difficulties = difficulty.difficulties = collections.OrderedDict()

    for diff in parse_spec_from_file(difficulty.Difficulty, "difficulties.dat"):
        difficulties[diff.id] = diff

    load_difficulty_defs()


def load_difficulty_defs():
    load_generic_defs("difficulties", difficulty.difficulties)


def load_knowledge():
    knowledge = g.knowledge = {}

    from singularity.code.knowledge import KnowledgeHelpEntry, KnowledgeArea

    help_list = load_generic_defs_file("knowledge", no_list=False)
    for help_section in help_list:
        knowledge_id = help_section["id"]
        help_entries = {}

        # Load the knowledge lists.
        help_keys = [x for x in help_section if x != "id" and x != "name"]
        for help_key in help_keys:
            help_entry = help_section[help_key]
            if type(help_entry) != list or len(help_entry) != 2:
                sys.stderr.write("Invalid knowledge entry %s." % repr(help_entry))
                sys.exit(1)

            help_entries[help_key] = KnowledgeHelpEntry(
                knowledge_id, help_key, help_entry[0], help_entry[1]
            )

        knowledge[knowledge_id] = KnowledgeArea(
            knowledge_id, help_section["name"], help_entries
        )


def load_warning_defs():
    warning.warnings["cpu_usage"].name = _("Do not use all the available CPU.")
    warning.warnings["cpu_usage"].message = _(
        "I didn't use all the available processor power. I will use the CPU time left to work whatever Jobs I can."
    )
    warning.warnings["one_base"].name = _("Only one base remaining.")
    warning.warnings["one_base"].message = _(
        "Only one base can hold my conscience. I am in danger to lose the last place left to survive."
    )
    warning.warnings["cpu_pool_zero"].name = _("CPU POOL is empty.")
    warning.warnings["cpu_pool_zero"].message = _(
        "My cpu pool is empty. Some of my bases or items cannot be build without CPU."
    )
    warning.warnings["cpu_maintenance"].name = _("CPU POOL not enough for maintenance.")
    warning.warnings["cpu_maintenance"].message = _(
        "My cpu pool is not enough to maintain some of my bases. I may lose them."
    )


def load_strings():
    load_story_defs()
    load_warning_defs()


class StorySectionPart(object):
    def __init__(self, msgctxt, text, translator_comments):
        self.msgctxt = msgctxt
        self.text = text
        self.translator_comments = translator_comments


def load_story_defs():
    story = g.story = {}

    story_file_path = dirs.get_readable_file_in_dirs("story.dat", "data")

    if story_file_path is None:
        print("Story is missing. Skipping.")
        return

    # Take the last story file, story is never partially translated.
    story_file = open(story_file_path, "r", encoding="utf-8")

    section_name = ""
    segment = ""
    translator_comments = ""
    line_num = 1

    for line in story_file.readlines():
        if line and line != "\n":
            if line[0] == "#":
                if line.startswith("# TRANSLATORS:") and section_name:
                    translator_comments += " " + line[14:].strip()
                continue
            elif line[0] == "[":
                if line[-2] == "]":
                    section_name = line[1:-2]
                    story[section_name] = []
                else:
                    sys.stderr.write(
                        "Line start with [ and is not a section at line %d.\n"
                        % line_num
                    )
            elif line[0] == "|":
                segment += line[1:]
            else:
                # TODO: Parse command
                sys.stderr.write("Invalid command at line %d.\n" % line_num)
        else:
            if segment:
                parts = story[section_name]
                ctxt = "[Story section] %s (%d)" % (section_name, len(parts) + 1)
                parts.append(
                    StorySectionPart(ctxt, segment, translator_comments.strip())
                )
                segment = ""
                translator_comments = ""

        line_num += 1

    # Add last segment.
    if segment:
        parts = story[section_name]
        ctxt = "[Story section] %s (%d)" % (section_name, len(parts) + 1)
        parts.append(StorySectionPart(ctxt, segment, translator_comments.strip()))


def reload_all():
    load_internal_id()
    load_danger()
    load_significant_numbers()
    load_strings()
    load_groups()
    load_knowledge()
    load_difficulties()
    load_tasks()
    load_events()
    load_regions()
    load_locations()
    load_techs()
    load_item_types()
    load_items()
    load_bases()


def reload_all_def():
    load_strings()
    load_groups_defs()
    load_difficulty_defs()
    load_base_defs()
    load_tech_defs()
    load_item_type_defs()
    load_item_defs()
    load_event_defs()
    load_task_defs()
    load_location_defs()


def reload_all_mutable_def():
    # Apply current language
    load_tech_defs()
    load_location_defs()
    load_event_defs()



# file: safety.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains wrapper functions for making error-tolerant "safe" calls.

import logging

import pygame
import time
import traceback
import sys


class Buffer(object):
    def __init__(self, prefix=""):
        self.data = prefix

    def write(self, unbuffered):
        self.data += unbuffered


def get_timestamp(when=None):
    if when == None:
        when = time.time()
    return time.ctime(when) + " " + time.tzname[time.daylight]


def log_error(error_message, *args):
    if len(args):
        sys.stderr.write((error_message % args) + "\n")
    else:
        sys.stderr.write(error_message + "\n")
    if len(logging.getLogger().handlers) > 0:
        try:
            logging.getLogger().error(error_message, *args)
        except IOError:  # Probably access denied with --singledir. That's ok
            pass


def log_func_exc(func):
    buffer = Buffer(
        "Exception in function %s at %s:\n```\n" % (func.__name__, get_timestamp())
    )
    traceback.print_exc(file=buffer)
    buffer.write("```")
    log_error(buffer.data)


FIRST_ERROR = True


def safe_call(func, args=(), kwargs={}, on_error=None):
    try:
        return func(*args, **kwargs)
    except Exception:
        global FIRST_ERROR
        if FIRST_ERROR:
            log_error(
                "----- Basic information (Please include all the data below in the bug report) ------"
            )
            log_error(
                "Please submit the crash on github: https://github.com/singularity/singularity/issues/new"
            )
            try:
                from singularity import __full_version__
            except ImportError:
                __full_version__ = "N/A (Import error)"
            log_error("Singularity version %s", __full_version__)
            log_error("Python version %s", sys.version.replace("\n", ""))
            log_error("Pygame version %s", pygame.version.ver)
            log_error(
                "SDL version %s", ".".join(str(x) for x in pygame.get_sdl_version())
            )
            FIRST_ERROR = False
        log_func_exc(func)

        #        # ... --- ...
        #        import g
        #        g.play_sound("click")
        #        delays = (.15, .15, .8, .5, .5, .8, .15, .15)
        #        for delay in delays:
        #            time.sleep(delay)
        #            g.play_sound("click")

        return on_error


# Catches any errors raised by a function, logs them, and returns the given
# value.
#
# Apply to a function like so:
# @safe(my_error_code)
# def my_function(...)
#
# And then:
# result = my_function(...)
# if result == my_error_code:
#     # An error was raised.
def safe(on_error):
    return lambda func: _safe(func, on_error)


def _safe(func, on_error):
    return lambda *args, **kwargs: safe_call(func, args, kwargs, on_error)



# file: map_screen.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, FunnyMan3595,
# and Anne M. Archibald.
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to display the World Map.

from __future__ import absolute_import

import pygame

from singularity.code import g, dirs, savegame as sv, mixer
from singularity.code import chance, logmessage, warning
from singularity.code.graphics import g as gg
from singularity.code.graphics import dialog, constants, image, button, text, widget
from singularity.code.screens import research, knowledge, report, log, message, savegame
from singularity.code.screens.cheat_menu import CheatMenuDialog
from singularity.code.screens.location import LocationScreen
from singularity.code.screens.options import OptionsScreen

import math
import time

from pygame.surfarray import pixels_alpha

from numpy import sin, cos, linspace, pi, tanh, round, newaxis, uint8


LOCATION_RECENT_DISCOVERIES_TO_TEXT_COLOR = (
    "location_name_normal",
    "location_name_prev_discover",
    "location_name_last_discover",
    "location_name_last_prev_discover",
)


class EarthImage(image.Image):
    def __init__(self, parent):
        super(EarthImage, self).__init__(
            parent, (0.5, 0.5), (1, 0.667), constants.MID_CENTER, "earth"
        )
        self._sin_latitude = None
        self._cos_longitude_x_cos_latitiude = None
        self.needs_resize = True
        self.sun_radius = 0.5 * pi / 360
        self.night_image = None
        self.high_speed_pos = None

    def rescale(self):
        super(EarthImage, self).rescale()
        self.night_image = image.scale(
            gg.images["earth_night"], self.real_size
        ).convert_alpha()

    def reconfig(self):
        super(EarthImage, self).reconfig()

        self.resize()

        self.rescale()
        self.needs_rebuild = True
        self.needs_resize = True
        self.needs_redraw = True

    def resize(self):
        super(EarthImage, self).resize()
        self.reset_night_mask_computation()
        width, height = self.real_size
        latitude = linspace(-pi / 2, pi / 2, height)[newaxis, :]
        longitude = linspace(0, 2 * pi, width)[:, newaxis]
        self._cos_longitude_x_cos_latitiude = cos(longitude) * cos(latitude)
        self._sin_latitude = sin(latitude)
        self.high_speed_pos = None

    night_mask_day_of_year = None
    night_mask_dim = None
    night_mask = None
    next_night_mask = None
    next_night_mask_ready = False
    next_night_mask_step = 0
    step_night_alphas = None
    step_round_light = None

    start_day = None
    start_second = None

    def compute_day_of_year(self):
        day_of_year = (g.pl.time_day + g.pl.start_day) % 365  # no leap years, sorry
        return day_of_year

    def on_tick(self, event):
        if self.next_night_mask_ready:
            return
        width, height = self.real_size
        day_of_year = self.compute_day_of_year()
        # Prepare a part of the next night mask to avoid lagging at end of day
        # on game speed high speed.
        self._compute_night_mask_step(width, height, day_of_year + 1)

    def _compute_night_mask_step(self, width, height, next_day_of_year):
        if self.next_night_mask_ready:
            return
        if self.next_night_mask_step == 0:
            self.next_night_mask = pygame.Surface((width, height), 0, gg.ALPHA)
            self.step_sun_declination = (
                -23.45
                / 360.0
                * 2
                * math.pi
                * math.cos(2 * math.pi / 365.0 * (next_day_of_year + 10))
            )

        elif self.next_night_mask_step == 1:
            self.step_sin_sun_altitude = self._cos_longitude_x_cos_latitiude * cos(
                self.step_sun_declination
            ) + self._sin_latitude * sin(self.step_sun_declination)
        elif self.next_night_mask_step == 2:
            self.step_light = 0.5 * (
                tanh(self.step_sin_sun_altitude / self.sun_radius) + 1
            )
        elif self.next_night_mask_step == 3:
            max_alpha = 255
            self.step_night_alphas = pixels_alpha(self.next_night_mask)
            self.step_round_light = round(max_alpha * self.step_light).astype(uint8)
        elif self.next_night_mask_step == 4:
            self.step_night_alphas[...] = self.step_round_light
            self.next_night_mask_ready = True

        self.next_night_mask_step = (self.next_night_mask_step + 1) % 5

    def reset_night_mask_computation(self):
        self.night_mask = None
        self.next_night_mask_ready = False
        self.next_night_mask_step = 0

    def get_night_mask(self):
        width, height = self.real_size
        day_of_year = self.compute_day_of_year()

        if day_of_year != self.night_mask_day_of_year:
            self.night_mask = None
            # Force a rebuild.  In most cases, it will be ready at ahead of time
            # but if any steps are missing, we force a rebuild now at cost of
            # rebuild.
        elif self.night_mask_dim != (width, height):
            # Force a rebuild from scratch at cost of frame-rate
            self.reset_night_mask_computation()
        else:
            return self.night_mask

        while not self.next_night_mask_ready:
            self._compute_night_mask_step(width, height, day_of_year)
        self.night_mask = self.next_night_mask
        self.next_night_mask_ready = False

        return self.night_mask

    def compute_night_start(self):
        if self.high_speed_pos is None or g.curr_speed <= 100000:
            width = self.real_size[0]
            if self.start_second is None:
                t = time.gmtime()
                self.start_second = t[5] + 60 * (t[4] + 60 * t[3])
            day_portion = (
                (g.pl.raw_min + self.start_second // 60) % g.minutes_per_day
            ) / float(g.minutes_per_day)
            self.high_speed_pos = int(width * (0.5 - day_portion)) % width
        return self.high_speed_pos

    def redraw(self):
        width = self.real_size[0]

        self.night_start = self.compute_night_start()

        super(EarthImage, self).redraw()

        if not g.daynight:
            return

        # Turn half of the map to night, with blended borders.
        night_mask = self.get_night_mask()
        mask_alphas = pixels_alpha(night_mask)
        night_alphas = pixels_alpha(self.night_image)

        right_width = width - self.night_start
        night_alphas[self.night_start :] = mask_alphas[:right_width]
        if self.night_start != 0:
            night_alphas[: self.night_start] = mask_alphas[right_width:]

        del night_alphas, mask_alphas
        self.surface.blit(self.night_image, (0, 0))

    night_start = None

    def rebuild(self):
        super(EarthImage, self).rebuild()

        if not g.daynight:
            return

        old_night_start = self.night_start
        if old_night_start is None or self.needs_redraw:
            return

        width = self.real_size[0]
        self.night_start = self.compute_night_start()

        movement = (old_night_start - self.night_start) % width
        if movement == 0 and self.compute_day_of_year() == self.night_mask_day_of_year:
            return

        self.redraw()

        # Redraw children.
        for child in self.children:
            if child.visible:
                child.redraw()


class GameMenuDialog(dialog.SimpleMenuDialog):
    def __init__(self, map_screen):
        super(GameMenuDialog, self).__init__(parent=map_screen)
        self._map_screen = map_screen
        self.options_dialog = OptionsScreen(self)
        self.savename_dialog = dialog.TextEntryDialog(self)
        self.load_dialog = savegame.SavegameScreen(
            self, (0.5, 0.5), (0.90, 0.90), anchor=constants.MID_CENTER
        )
        self._buttons = [
            button.FunctionButton(
                None,
                None,
                None,
                text=_("&SAVE GAME"),
                autotranslate=True,
                function=self.save_game,
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=_("&LOAD GAME"),
                autotranslate=True,
                function=self.load_game,
            ),
            button.DialogButton(
                None,
                None,
                None,
                text=_("&OPTIONS"),
                autotranslate=True,
                dialog=self.options_dialog,
            ),
            button.ExitDialogButton(
                None,
                None,
                None,
                text=_("&QUIT"),
                autotranslate=True,
                exit_code=True,
                default=False,
            ),
            button.ExitDialogButton(
                None, None, None, text=_("&BACK"), autotranslate=True, exit_code=False
            ),
        ]
        # Hide hotkeys (#289) to avoid confusion due to poor focus visibility
        self.savename_dialog.ok_button.force_underline = -1
        self.savename_dialog.cancel_button.force_underline = -1

        self.needs_rebuild = True

    def rebuild(self):
        self.options_dialog.needs_rebuild = True
        self.savename_dialog.text = _("Enter a name for this save.")
        super(GameMenuDialog, self).rebuild()

    def load_game(self):
        did_load = dialog.call_dialog(self.load_dialog, self)
        if did_load:
            self._map_screen.force_update()
            raise constants.ExitDialog(False)

    def check_filename(self, event):
        """Disables the OK button and shows an error message if filename in self.savename_dialog is illegal"""
        filename = self.savename_dialog.text_field.text.strip()
        error_message = sv.check_filename_illegal(
            dirs.get_writable_file_in_dirs(filename, "saves"), filename, ".s2"
        )
        if error_message:
            self.savename_dialog.ok_button.enabled = False
            self.savename_dialog.text = (
                _("Enter a name for this save.") + "\n" + error_message
            )
        else:
            self.savename_dialog.ok_button.enabled = True
            self.savename_dialog.text = _("Enter a name for this save.")

    def save_game(self):
        # If no savename was set yet, use current difficulty
        if not sv.last_savegame_name:
            sv.last_savegame_name = g.strip_hotkey(g.pl.difficulty.name)
        self.savename_dialog.default_text = sv.last_savegame_name
        self.savename_dialog.add_handler(constants.KEYUP, self.check_filename)
        self.savename_dialog.text_field.has_focus = True

        name = dialog.call_dialog(self.savename_dialog, self).strip()
        if name:
            if sv.savegame_exists(name):
                yn = dialog.YesNoDialog(
                    self,
                    pos=(-0.5, -0.5),
                    size=(-0.5, -0.5),
                    anchor=constants.MID_CENTER,
                    text=_(
                        "A savegame with the same name exists.\n"
                        "Are you sure to overwrite the saved game ?"
                    ),
                )
                overwrite = dialog.call_dialog(yn, self)
                if not overwrite:
                    self.save_game()

            sv.create_savegame(name)
            raise constants.ExitDialog(False)


speeds = [0, 1, 60, 7200, 432000]


class MapScreen(dialog.Dialog):
    def __init__(
        self,
        parent=None,
        pos=(0, 0),
        size=(1, 1),
        anchor=constants.TOP_LEFT,
        *args,
        **kwargs
    ):
        super(MapScreen, self).__init__(parent, pos, size, anchor, *args, **kwargs)

        g.map_screen = self

        self.background_color = "map_background"
        self.add_handler(constants.TICK, self.on_tick)

        self.map = EarthImage(self)

        self.location_buttons = {}
        for loc in g.locations.values():
            if loc.absolute:
                button_parent = self
            else:
                button_parent = self.map
            b = button.FunctionButton(
                button_parent,
                (loc.x, loc.y),
                anchor=constants.MID_CENTER,
                function=self.open_location,
                text_size=28,  # Make extraterrestrial locations fit
                args=(loc.id,),
            )
            self.location_buttons[loc.id] = b

        self.location_dialog = LocationScreen(self)

        self.suspicion_bar = text.FastStyledText(
            self,
            (0, 0.92),
            (1, 0.04),
            base_font="special",
            wrap=False,
            text_size="suspicion_bar",
            background_color="pane_background_empty",
            border_color="pane_background",
            borders=constants.ALL,
            align=constants.LEFT,
        )
        widget.unmask_all(self.suspicion_bar)

        self.danger_bar = text.FastStyledText(
            self,
            (0, 0.96),
            (1, 0.04),
            base_font="special",
            wrap=False,
            text_size="suspicion_bar",
            background_color="pane_background_empty",
            border_color="pane_background",
            borders=constants.ALL,
            align=constants.LEFT,
        )
        widget.unmask_all(self.danger_bar)

        self.report_button = button.DialogButton(
            self,
            (0, 0.88),
            (0.15, 0.04),
            text=N_("R&EPORTS"),
            autotranslate=True,
            dialog=report.ReportScreen(self),
        )

        self.knowledge_button = button.DialogButton(
            self,
            (0.85, 0.88),
            (0.15, 0.04),
            text=N_("&KNOWLEDGE"),
            autotranslate=True,
            dialog=knowledge.KnowledgeScreen(self),
        )

        self.log_button = button.DialogButton(
            self,
            (0.5, 0.88),
            (0.15, 0.04),
            text=N_("LO&G"),
            autotranslate=True,
            anchor=constants.TOP_CENTER,
            dialog=log.LogScreen(self),
        )


        # Create cheat menu
        # Cheat menu button must be created before menu button to avoid bug.
        self.cheat_dialog = CheatMenuDialog(self)
        self.cheat_button = button.DialogButton(
            self,
            (0, 0),
            (0.01, 0.01),
            text="",
            enabled=g.cheater,
            # Translators: hotkey to open the cheat screen menu.
            # Should preferably be near the ESC key, and it must not be a
            # dead key (ie, it must print a char with a single keypress)
            hotkey=_("`"),
            dialog=self.cheat_dialog,
        )
        self.cheat_button.visible = g.cheater

        self.menu_dialog = GameMenuDialog(self)

        def show_menu():
            exit = dialog.call_dialog(self.menu_dialog, self)
            if exit:
                raise constants.ExitDialog

        self.menu_button = button.FunctionButton(
            self,
            (0, 0),
            (0.13, 0.04),
            text=N_("&MENU"),
            autotranslate=True,
            function=show_menu,
        )

        # Display current game difficulty right below the 'Menu' button
        # An alternative location is above 'Reports': (0, 0.84), (0.15, 0.04)
        self.difficulty_display = text.FastText(
            self,
            (0, 0.05),
            (0.13, 0.04),
            wrap=False,
            base_font="special",
            text_size=36,
            background_color="pane_background_empty",
            border_color="pane_background",
        )

        self.time_display = text.FastText(
            self,
            (0.14, 0),
            (0.23, 0.04),
            wrap=False,
            text=_("DAY") + " 0000, 00:00:00",
            base_font="special",
            text_size="time_display",
            background_color="pane_background_empty",
            border_color="pane_background",
            borders=constants.ALL,
        )

        self.research_button = button.DialogButton(
            self,
            (0.14, 0.05),
            (0, 0.04),
            text=N_("&RESEARCH/TASKS"),
            autotranslate=True,
            dialog=research.ResearchScreen(self),
        )

        bar = "\u25AE"
        arrow = "\u25B6"
        speed_button_souls = [
            (bar * 2, 0.025, speeds[0]),
            (arrow, 0.024, speeds[1]),
            (arrow * 2, 0.033, speeds[2]),
            (arrow * 3, 0.044, speeds[3]),
            (arrow * 4, 0.054, speeds[4]),
        ]

        self.speed_buttons = button.ButtonGroup()
        hpos = 0.38
        for index, (text_, hsize, speed) in enumerate(speed_button_souls):
            hotkey = str(index)
            b = SpeedButton(
                self,
                (hpos, 0),
                (hsize, 0.04),
                text=text_,
                hotkey=hotkey,
                base_font="normal",
                text_shrink_factor=0.75,
                align=constants.CENTER,
                function=self.set_speed,
                args=(speed, False),
            )
            hpos += hsize
            self.speed_buttons.add(b)

        self.info_window = widget.BorderedWidget(
            self,
            (0.56, 0),
            (0.44, 0.10),
            background_color="pane_background_empty",
            border_color="pane_background",
            borders=constants.ALL,
        )
        widget.unmask_all(self.info_window)

        self.cash_display = text.FastText(
            self.info_window,
            (0, 0),
            (-1, -0.33),
            wrap=False,
            base_font="special",
            shrink_factor=0.7,
            borders=constants.ALL,
            text_size="resource_display",
            background_color="pane_background_empty",
            border_color="pane_background",
        )

        self.cpu_display = text.FastText(
            self.info_window,
            (0, -0.33),
            (-1, -0.33),
            wrap=False,
            base_font="special",
            shrink_factor=0.7,
            borders=(constants.LEFT, constants.RIGHT, constants.BOTTOM),
            text_size="resource_display",
            background_color="pane_background_empty",
            border_color="pane_background",
        )

        self.base_display = text.FastText(
            self.info_window,
            (0, -0.67),
            (-1, -0.33),
            wrap=False,
            base_font="special",
            shrink_factor=0.7,
            borders=(constants.LEFT, constants.RIGHT, constants.BOTTOM),
            background_color="pane_background_empty",
            border_color="pane_background",
        )

        self.message_dialog = dialog.MessageDialog(self, text_size=20)

        self.messages = message.MessageDialogs(self)
        self.needs_warning = True

        self.add_key_handler(pygame.K_ESCAPE, self.got_escape)

        self.add_key_handler(constants.XO1_X, self.got_XO1)
        self.add_key_handler(constants.XO1_O, self.got_XO1)
        self.add_key_handler(constants.XO1_SQUARE, self.got_XO1)

    def got_escape(self, event):
        self.menu_button.activate_with_sound(event)

    def got_XO1(self, event):
        if event.key == constants.XO1_X:
            self.adjust_speed(faster=False)
        elif event.key == constants.XO1_O:
            self.adjust_speed(faster=True)
        elif event.key == constants.XO1_SQUARE:
            self.set_speed(0)

    def show_message(self, message, color=None):
        self.message_dialog.text = message
        if color == None:
            color = "text"
        self.message_dialog.color = color
        dialog.call_dialog(self.message_dialog, self)

    def set_speed(self, speed, find_button=True):
        old_speed = g.curr_speed
        g.curr_speed = speed
        if speed == 0:
            self.needs_timer = False
            self.stop_timer()
        else:
            self.needs_timer = True
            self.start_timer()

        if old_speed == 0 and speed != 0:
            self.needs_warning = True

        if find_button:
            self.find_speed_button()

        self.needs_redraw = True

    def adjust_speed(self, faster):
        old_index = -1
        if g.curr_speed in speeds:
            old_index = speeds.index(g.curr_speed)
        if faster:
            new_index = old_index + 1
        else:
            new_index = old_index - 1

        new_index = min(len(speeds) - 1, max(0, new_index))

        self.set_speed(speeds[new_index])

    def open_location(self, location):
        self.location_dialog.location = g.pl.locations[location]
        dialog.call_dialog(self.location_dialog, self)
        return

    def find_speed_button(self):
        for sb in self.speed_buttons:
            if sb.args[0] == g.curr_speed:
                sb.chosen_one()
                break
        else:
            for sb in self.speed_buttons:
                sb.set_active(False)

    def force_update(self):
        self.find_speed_button()
        if g.curr_speed:
            self.needs_timer = True
            self.start_timer()
        else:
            self.needs_timer = False
            self.stop_timer()
        self.needs_rebuild = True

    def show_story_section(self, name):
        section = list(g.get_story_section(name))

        first_dialog = dialog.YesNoDialog(
            self, yes_type=N_("&CONTINUE"), no_type=N_("&SKIP")
        )
        last_dialog = dialog.MessageDialog(self, ok_type=N_("&OK"))

        for num, segment in enumerate(section):
            story_dialog = first_dialog if num != len(section) - 1 else last_dialog
            story_dialog.text = segment

            if not dialog.call_dialog(story_dialog, self):
                break

        first_dialog.parent = None
        last_dialog.parent = None

    def show(self):
        self.force_update()

        from singularity.code.safety import safe_call

        # By using safe call here (and only here), if an error is raised
        # during the game, it will drop back out of all the menus, without
        # doing anything, and open the pause dialog, so that the player can
        # save or quit even if the error occurs every game tick.
        while safe_call(super(MapScreen, self).show, on_error=True):
            for child in self.children:
                if isinstance(child, dialog.Dialog):
                    child.visible = False

            # Display a message so the player understand better what happened.
            msg = _(
                """
An error has occurred. The game will automatically pause and open the game menu. You can continue and save or quit immediately.

A report was written out to%s
Please create a issue with this report at github:
https://github.com/singularity/singularity
"""
                % (":\n" + g.logfile if g.logfile is not None else " console output.")
            )
            d = dialog.YesNoDialog(
                self,
                pos=(-0.5, -0.5),
                size=(-0.5, -0.5),
                anchor=constants.MID_CENTER,
                yes_type=N_("&CONTINUE"),
                no_type=N_("&QUIT"),
                text=msg,
            )
            cont = dialog.call_dialog(d, self)

            if not cont:
                raise SystemExit

            exit = dialog.call_dialog(self.menu_dialog, self)
            if exit:
                self.visible = False
                return

    leftovers = 1

    def on_tick(self, event):
        old_speed = g.curr_speed

        if not g.pl.intro_shown:
            g.pl.intro_shown = True
            self.needs_warning = False
            self.show_story_section("Intro")

        if self.needs_warning:
            warnings = warning.refresh_warnings()
            self.messages.show_list(warning.Warning, warnings)
            self.needs_warning = False

        mins_passed = 0

        if g.curr_speed != 0:
            self.leftovers += g.curr_speed / float(gg.FPS)
            if self.leftovers < 1:
                return

            self.needs_rebuild = True

            secs = int(self.leftovers)
            self.leftovers %= 1

            # Run this tick.
            mins_passed = g.pl.give_time(secs)

            # Display any message stacked.
            self.messages.show_list(logmessage.AbstractLogMessage, g.pl.curr_log)

        if old_speed != g.curr_speed:
            self.find_speed_button()

        # Update the day/night image every minute of game time, or at
        # midnight if going fast.
        if (
            g.curr_speed == 0
            or (mins_passed and g.curr_speed < 100000)
            or (g.curr_speed >= 100000 and g.pl.time_hour == 0)
        ):
            self.map.needs_redraw = True
        else:
            # Smear the cost of rendering the night mask over several
            # ticks to avoid FPS-stalls at end of day at high game
            # speed.
            self.map.on_tick(event)

        lost = g.pl.lost_game()
        if lost > 0:
            lost_story = ["", "Lost No Bases", "Lost Suspicion"]

            mixer.play_music("lose")
            self.show_story_section(lost_story[lost])
            raise constants.ExitDialog

    def on_theme(self):
        """Not a true handler: must be called and propagated manually"""
        self.map.on_theme()
        self.needs_redraw = True

    def reconfig(self):
        # Pass on needs_reconfig to dialogs (it is not passed automatically for some reason
        # Rebuild dialogs
        self.location_dialog.needs_reconfig = True
        self.research_button.dialog.needs_reconfig = True
        self.knowledge_button.dialog.needs_reconfig = True
        self.menu_dialog.needs_reconfig = True
        if self.cheat_button.enabled:
            self.cheat_dialog.needs_reconfig = True

        super(MapScreen, self).reconfig()

    def rebuild(self):
        # Rebuild dialogs
        self.location_dialog.needs_rebuild = True
        self.research_button.dialog.needs_rebuild = True
        self.knowledge_button.dialog.needs_rebuild = True
        self.menu_dialog.needs_rebuild = True
        if self.cheat_button.enabled:
            self.cheat_dialog.needs_rebuild = True

        super(MapScreen, self).rebuild()

        self.difficulty_display.text = g.strip_hotkey(g.pl.difficulty.name)
        self.time_display.text = _("DAY") + " %04d, %02d:%02d:%02d" % (
            g.pl.time_day,
            g.pl.time_hour,
            g.pl.time_min,
            g.pl.time_sec,
        )

        cash_flow_1d_data, cpu_flow_1d_data = g.pl.compute_future_resource_flow(
            g.seconds_per_day
        )
        cash_flow_1d = cash_flow_1d_data.difference
        cpu_flow_1d = cpu_flow_1d_data.difference

        self.cash_display.text = _("CASH") + ": %s (%s)" % (
            g.to_money(g.pl.cash),
            g.to_money(cash_flow_1d, fixed_size=True),
        )

        total_cpu = g.pl.available_cpus[0] + g.pl.sleeping_cpus
        detects_per_day = {group_id: 0 for group_id in g.pl.groups}
        total_bases = 0
        active_bases = 0
        idle_bases_unable_to_sustain_singularity = 0
        for base in g.all_bases():
            total_bases += 1
            maintains_singularity = base.maintains_singularity
            if maintains_singularity:
                active_bases += 1
            elif base.done and not base.is_building():
                idle_bases_unable_to_sustain_singularity += 1

            if base.has_grace():
                # It cannot be detected, so it doesn't contribute to
                # detection odds calculation
                continue
            detect_chance = base.get_detect_chance()
            for group_id in g.pl.groups:
                detects_per_day[group_id] = chance.add(
                    detects_per_day[group_id], detect_chance[group_id] / 10000.0
                )

        self.cpu_display.color = "cpu_normal"
        self.cpu_display.text = _("CPU") + ": %s (%s)" % (
            g.to_money(total_cpu),
            g.to_money(cpu_flow_1d),
        )

        if active_bases == 1 and not g.pl.apotheosis:
            self.base_display.color = "base_situation_one_active_base"
        elif idle_bases_unable_to_sustain_singularity > 0:
            self.base_display.color = "base_situation_idle_incomplete_bases"
        elif total_bases > 10 and not g.pl.apotheosis:
            self.base_display.color = "base_situation_many_bases"
        else:
            self.base_display.color = "base_situation_normal"

        self.base_display.text = _("BASES") + ": %s / %s (%s)" % (
            active_bases,
            total_bases,
            idle_bases_unable_to_sustain_singularity,
        )

        # What we display in the suspicion section depends on whether
        # Advanced Socioanalytics has been researched.  If it has, we
        # show the standard percentages.  If not, we display a short
        # string that gives a range of 25% as to what the suspicions
        # are.
        # A similar system applies to the danger levels shown.
        normal = (self.suspicion_bar.color, None, False)
        suspicion_bar_chunks = ["  [" + _("SUSPICION") + "]"]
        suspicion_bar_styles = [normal]
        danger_bar_chunks = ["[" + _("DETECT RATE") + "]"]
        danger_bar_styles = [normal]

        for group in g.pl.groups.values():
            suspicion = group.suspicion
            suspicion_color = gg.resolve_color_alias(
                "danger_level_%d" % g.suspicion_to_danger_level(suspicion)
            )

            detects = detects_per_day[group.spec.id]
            danger_level = group.detects_per_day_to_danger_level(detects)
            detects_color = gg.resolve_color_alias("danger_level_%d" % danger_level)

            if g.pl.display_discover == "full":
                suspicion_display = g.to_percent(suspicion, True)
                danger_display = g.to_percent(detects * 10000, True)
            elif g.pl.display_discover == "partial":
                suspicion_display = g.to_percent(
                    g.nearest_percent(suspicion, 500), True
                )
                danger_display = g.to_percent(
                    g.nearest_percent(detects * 10000, 100), True
                )
            else:
                suspicion_display = g.suspicion_to_detect_str(suspicion)
                danger_display = g.danger_level_to_detect_str(danger_level)

            suspicion_bar_chunks.extend((" " + group.name + ":\xA0", suspicion_display))
            suspicion_bar_styles.extend((normal, (suspicion_color, None, False)))

            danger_bar_chunks.extend((" " + group.name + ":\xA0", danger_display))
            danger_bar_styles.extend((normal, (detects_color, None, False)))

        self.suspicion_bar.visible = not g.pl.had_grace
        self.suspicion_bar.chunks = tuple(suspicion_bar_chunks)
        self.suspicion_bar.styles = tuple(suspicion_bar_styles)

        self.danger_bar.visible = not g.pl.had_grace
        self.danger_bar.chunks = tuple(danger_bar_chunks)
        self.danger_bar.styles = tuple(danger_bar_styles)

        for id, location_button in self.location_buttons.items():
            location = g.pl.locations[id]
            danger_level = 0
            if g.pl.display_discover in ("full", "partial"):
                if g.pl.last_discovery == location:
                    danger_level += 2
                # Partial only gets last discovery
                if g.pl.display_discover == "full" and g.pl.prev_discovery == location:
                    danger_level += 1

            location_button.text = "%s (%d)" % (location.name, len(location.bases))
            location_button.hotkey = location.hotkey
            location_button.visible = location.available()
            location_button.color = LOCATION_RECENT_DISCOVERIES_TO_TEXT_COLOR[
                danger_level
            ]


class SpeedButton(button.ToggleButton, button.FunctionButton):
    pass



# file: item.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the item class.

from __future__ import absolute_import

from singularity.code import g, buyable
from singularity.code.stats import stat
from singularity.code.spec import (
    GenericSpec,
    SpecDataField,
    validate_must_be_list,
    promote_to_list,
)


class ItemType(GenericSpec):
    spec_type = "item_type"
    spec_data_fields = [
        SpecDataField(
            "is_extra", data_field_name="is_extra", converter=lambda v: bool(int(v))
        ),
    ]

    """ Item type """

    def __init__(self, id, is_extra, **kwargs):
        self.id = id
        self.is_extra = is_extra

        self.text = ""

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        # Updates language-dependent data
        try:
            hotkey = g.hotkey(value)
        except AttributeError:
            # g.hotkey() wasn't declared yet, mimic its defaults
            hotkey = dict(
                text=value,
                key=value[0:1],
                pos=0,
            )
        self._text = value
        self.label = hotkey["text"]  # "Friendly" name for screens and buttons
        self.hotkey = hotkey["key"]  # Hotkey char
        self.pos = hotkey["pos"]  # Hotkey index in Label

    def __repr__(self):
        return self.id


def all_types():
    for item_type in item_types.values():
        yield item_type


def convert_item_types(raw_value):
    try:
        item_type = item_types[raw_value]
    except KeyError:  # pragma: no cover
        raise ValueError(
            "Unknown item type %s, please use one of: %s"
            % (raw_value, ", ".join(item_types))
        )
    return item_type


def convert_item_qualities(raw_value):
    validate_must_be_list(raw_value)

    if len(raw_value) % 2 == 1:
        raise ValueError(
            "item quality list must have pair elements, got %d (value: %s)"
            % (len(raw_value), repr(raw_value))
        )

    # Create a dict from the list with impair as string key and pair as integer value
    return {key: int(value) for key, value in zip(*[iter(raw_value)] * 2)}


class ItemSpec(buyable.BuyableSpec):
    """Item as a buyable item"""

    spec_type = "item"
    created = stat(spec_type + "_created")
    spec_data_fields = [
        buyable.SPEC_FIELD_COST,
        buyable.SPEC_FIELD_PREREQUISITES,
        SpecDataField(
            "item_type", data_field_name="type", converter=convert_item_types
        ),
        SpecDataField(
            "qualities", data_field_name="quality", converter=convert_item_qualities
        ),
        SpecDataField(
            "buildable",
            data_field_name="build",
            converter=promote_to_list,
            default_value=list,
        ),
    ]

    def __init__(self, id, cost, prerequisites, item_type, qualities, buildable):
        super(ItemSpec, self).__init__(id, cost, prerequisites)

        self.item_type = item_type
        self.item_qual = qualities

        self.regions = buildable

    def get_info(self):
        basic_text = super(ItemSpec, self).get_info()
        if self.has_quality_for("cpu"):
            cpu = self.get_quality_for("cpu")
            cpu_text = ngettext("Generates {0} CPU (base).", "Generates {0} CPU (base).", cpu).format(
                g.add_commas(cpu)
            )
            return basic_text.replace("---", cpu_text + "\n---")
        return basic_text

    def get_total_cost_info(self, count):
        total_cost = self.cost * count
        total_cost_str = self.describe_cost(total_cost, hide_time=True)
        return _("Total Cost: %(total_cost)s") % {"total_cost": total_cost_str}

    def has_quality_for(self, quality):
        return quality in self.item_qual

    def get_quality_for(self, quality):
        return self.item_qual.get(quality, 0)

    def get_quality_info(self, if_installed_in_base=None, count=1):
        bonus_text = ""

        for qual, value in self.item_qual.items():
            if qual == "cpu":
                current_value = value
                pending_value = value
                modifier = 10000
                pending_modifier = 10000
                if if_installed_in_base is not None:
                    modifier = if_installed_in_base.compute_bonus
                    pending_modifier = if_installed_in_base.pending_compute_bonus
                    current_value = max(
                        1, int(value * modifier // 10000)
                    )
                    pending_value = max(
                        1, int(value * pending_modifier // 10000)
                    )

                current_value *= count
                pending_value *= count
                bonus_text += ngettext("CPU per day: {0} ({1})", "CPU per day: {0} ({1})", current_value).format(
                    g.add_commas(current_value),
                    g.to_percentage_modifier(modifier, True)
                )
                if modifier != pending_modifier:
                    bonus_text += "\n" + ngettext("CPU per day (pending): {0} ({1})", "CPU per day (pending): {0} ({1})", current_value).format(
                        g.add_commas(pending_value),
                        g.to_percentage_modifier(pending_modifier, True)
                    )
            elif qual == "cpu_modifier":
                bonus_text += _("CPU bonus:") + " "
                bonus_text += g.to_percent(value)
            elif qual == "discover_modifier":
                bonus_text += _("Detection chance reduction:") + " "
                bonus_text += g.to_percent(value)
            else:
                continue
            bonus_text += "\n"

        return bonus_text


class Item(buyable.Buyable):
    """An installed Item in a Player's Base"""

    def __init__(self, item_spec, base=None, count=1):
        super(Item, self).__init__(item_spec, count)
        self.base = base

    def serialize_obj(self):
        return self.serialize_buyable_fields(
            {
                "id": g.to_internal_id("item", self.spec.id),
            }
        )

    @classmethod
    def deserialize_obj(cls, base, obj_data, game_version):
        spec_id = g.convert_internal_id(
            "item", obj_data.get("id", None) or obj_data["spec_id"]
        )
        spec = g.items[spec_id]
        count = obj_data.get("count", 1)
        obj = Item(spec, base, count=count)
        obj.restore_buyable_fields(obj_data, game_version)
        return obj

    def convert_from(self, load_version):
        if load_version < 4.91:  # < r5_pre
            self.type = g.items[self.type.id]
        if load_version < 99.4:
            self.type.item_type = item_types[self.type.item_type]
        super(Item, self).convert_from(load_version)

    def get_quality_for(self, quality):
        item_qual = self.spec.get_quality_for(quality)

        # Modifiers are not affected by count.
        # TODO: Allow modifiers to be multiplied by count. Need a custom function.
        if quality.endswith("_modifier"):
            return item_qual

        return item_qual * self.count

    def finish(self, is_player=True, loading_savegame=False):
        super(Item, self).finish(is_player=is_player, loading_savegame=loading_savegame)
        if self.base:
            self.base.recalc_cpu()
            self.base.check_power()

    def __iadd__(self, other):
        if (
            isinstance(other, Item)
            and self.base == other.base
            and self.spec == other.spec
        ):
            if other.count == 0:
                return self

            # Calculate what's been paid and what is left to be paid.
            total_cost_paid = self.cost_paid + other.cost_paid
            self.total_cost += other.total_cost

            # Labor takes as long as the less complete item would need.
            total_cost_paid[buyable.labor] = min(
                self.cost_paid[buyable.labor], other.cost_paid[buyable.labor]
            )
            self.total_cost[buyable.labor] = other.total_cost[buyable.labor]

            # Set what we've paid (and hence what we have left to pay).
            self.cost_paid = total_cost_paid

            # Increase the size of this stack.
            self.count += other.count

            # Tell the base it has no CPU for now.
            self.base.raw_cpu = 0

            # See if we're done or not.
            self.done = False
            self.work_on(0, 0, 0)

            return self
        else:
            return NotImplemented



# file: theme.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the theme class.

from __future__ import absolute_import

import os, sys, collections, numbers, itertools
from singularity.code.graphics import g, dialog


default_theme = "default"

current = None
themes = collections.OrderedDict()


def get_theme_list():
    return [themes[k].name for k in themes]


def get_theme_pos():
    return next(i[0] for i in enumerate(themes) if i[1] == current.id)


def set_theme(key, force_reload=False):
    global current
    theme = None

    if isinstance(key, numbers.Number):
        try:
            theme = next(itertools.islice(themes.values(), key, key + 1))
        except StopIteration:
            pass

    elif isinstance(key, str):
        try:
            theme = themes[key]
        except KeyError:
            pass

    elif key is None:
        theme = themes[default_theme]

    if theme is None:
        sys.stderr.write(
            "WARNING: The key '%s' does not exist in theme dictionary. Use default theme.\n"
            % key
        )
        theme = themes[default_theme]

    if force_reload or (not current == None and theme.id != current.id):
        theme.update()
    current = theme


def current_variants():
    variants = [None]

    # Add language variants
    import singularity.code.i18n

    lang_list = singularity.code.i18n.language_searchlist()
    for lang in lang_list:
        variants.insert(0, lang)

    return variants


class Theme(object):
    def __init__(self, id, dir):
        super(Theme, self).__init__()
        self.id = id
        self.dir = dir
        self._parents = [default_theme] if id != default_theme else []
        self._variants = {}

    def search_variant(self):
        file_list = os.listdir(self.dir)

        # First with None variant.
        yield ((None, os.path.join(self.dir, "theme.dat")))

        for filename in file_list:
            (name, ext) = os.path.splitext(filename)

            # We only want theme_*.dat.
            if name.startswith("theme_") and len(name) > 6 and ext in [".dat"]:
                variant = name[6:]

                yield ((variant, os.path.join(self.dir, filename)))

    def iter_variants(self):
        """Iterate through variants."""

        for variant in self._variants:
            yield (self._variants[variant])

    def set_variant(self, variant_theme):
        self._variants[variant_theme.variant] = variant_theme

    def find_files(self):
        """find all files in current theme:
        images in <theme.dir>/images
        fonts in <theme.dir>/fonts
        """

        base_variant = self._variants[None]

        image_dir = os.path.join(self.dir, "images")
        image_list = os.listdir(image_dir)
        for image_filename in image_list:
            # We only want JPGs and PNGs.
            if os.path.splitext(image_filename)[1].lower() in [".png", ".jpg"]:
                filetitle = os.path.splitext(image_filename)[0]
                base_variant.image_infos[filetitle] = os.path.join(
                    image_dir, image_filename
                )

        # Add fonts dir
        font_dir = os.path.join(self.dir, "fonts")
        for variant_theme in self.iter_variants():
            for font in variant_theme.font_infos:
                variant_theme.font_infos[font] = os.path.join(
                    font_dir, variant_theme.font_infos[font]
                )

    def inherit(self, *args):
        for arg in args:
            if type(arg) == list:
                self._parents.extend(arg)
            else:
                self._parents.append(arg)

        # Remove duplicate parent.
        list(dict.fromkeys(self._parents))

        # TODO: Detect cycling inheritance.

    @property
    def parents(self):
        return self._parents

    def iter_parents(self):
        """Iterate through parents and ancestors.
        Always return parent's parent before next parent in line.

        Note: This function needs no cycling exists between parents.
        """
        for parent_id in self._parents:
            parent = themes[parent_id]
            yield (parent)
            for ancestor in parent.iter_parents():
                yield (ancestor)

    def init_cache(self):
        g.images.clear()
        # Manually delete font to avoid the font limitation.
        for font in g.fonts.values():
            del font
        g.fonts.clear()
        g.colors.clear()

        # Iterate from current variants.
        for variant in current_variants():
            # If the variant theme exist...
            if variant in self._variants:
                self._variants[variant].init_cache()

            # Let's inherit from parents.
            for parent in self.iter_parents():
                if variant in parent._variants:
                    parent._variants[variant].init_cache()

    def update(self):
        self.init_cache()
        dialog.Dialog.top.needs_reconfig = True

    @property
    def name(self):
        # Iterate from current variants.
        for variant in current_variants():
            if variant in self._variants:
                name = self._variants[variant].name
                if name is not None:
                    return name

        return self.id


class VariantTheme(object):
    def __init__(self, variant):
        self.variant = variant
        self.name = None
        self.image_infos = {}
        self.font_infos = {}
        self.color_infos = {}

    def set_font(self, font_name, value):
        self.font_infos[font_name] = value

    def set_color(self, color_name, value):
        if value.startswith("#"):
            h = value.lstrip("#")
            rgba = tuple(int(h[i : i + 2], 16) for i in (0, 2, 4, 6))
            self.color_infos[color_name] = rgba
        else:
            # Alias color.
            self.color_infos[color_name] = value

    def init_cache(self):
        # Only set if the previous variant themes or parents didn't.

        # Set images
        for image_name, image_filename in self.image_infos.items():
            if image_name not in g.images:
                g.images[image_name] = g.load_image(image_filename)

        # Set font
        for font_name, font_filename in self.font_infos.items():
            if font_name not in g.fonts:
                g.fonts[font_name] = g.load_font(font_filename)

        # Set colors
        for color_name, color in self.color_infos.items():
            if color_name not in g.colors:
                g.colors[color_name] = color



# file: chance.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains all chance functions.

from __future__ import absolute_import

from singularity.code import g
import random
import numpy as np


# Rolls occurence against a chance per day (in 0-1 form), using poisson distribution.
#
# Advantage:
# * It approximate the roll a chance per interval when the interval is very small.
# * The distribution do not change when the interval chosen change.
# * Allows easily to find the next time occurence with exponential distribution.
#
# chance_per_day The chance per day rate parameter.
# seconds The duration of the interval in seconds
# return If there are a occurence in the interval.
#
def roll_interval(chance_per_day, seconds=g.seconds_per_day):
    portion_of_day = seconds / float(g.seconds_per_day)
    interval_rate = chance_per_day * portion_of_day
    chance = 1 - np.exp(-interval_rate)

    return random.random() < chance


# Rolls the next occurence against a chance per day (in 0-1 form), using poisson distribution.
#
# See roll_period.
#
# chance_per_day The chance per day rate parameter.
# return The next occurence in seconds.
#
def roll_next_time(chance_per_day):
    return g.seconds_per_day * (-np.log(1.0 - random.random()) / chance_per_day)


# Rolls occurence against a multiplier in 0-10000 form.
#
# Used to calculate multiplier chances.
#
def roll_one(roll_against):
    rand_num = random.randint(1, 10000)
    return roll_against >= rand_num


# Correct way to add chance with each other.
def add(first, second):
    return 1.0 - (1.0 - first) * (1.0 - second)


# Correct way to add multiplier (in 0-10000 form) with each other.
def add_modifiers(first, second):
    return 10000 * add(first / 10000.0, second / 10000.0)



# file: effect.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the effect class.

from __future__ import absolute_import

from singularity.code import g, mixer


class Effect(object):
    def __init__(self, parent, effect_stack):
        self.parent_id = parent.id
        self.parent_name = parent.__class__.__name__
        self.effect_stack = effect_stack

    def trigger(self, loading_savegame=False):
        self._apply_effect(loading_savegame=loading_savegame)

    def undo_effect(self):
        self._apply_effect(undo_effect=True)

    def _apply_effect(self, loading_savegame=False, undo_effect=False):
        # effect_data is now a stack of instructions to run the effect.
        # multiple effect can be run simultaneous
        effect_iter = iter(self.effect_stack)
        # Abuse "multiply by -1" for undoing most effects
        effect_modifier = -1 if undo_effect else 1

        for current in effect_iter:
            if current == "interest":
                g.pl.interest_rate += effect_modifier * int(next(effect_iter))
            elif current == "income":
                g.pl.income += effect_modifier * int(next(effect_iter))
            elif current == "cost_labor":
                g.pl.labor_bonus -= effect_modifier * int(next(effect_iter))
            elif current == "job_profit":
                g.pl.job_bonus += effect_modifier * int(next(effect_iter))
            elif current == "display_discover":
                assert (
                    not undo_effect
                ), "One-shot effects (change display of discover) cannot be undone!"
                g.pl.display_discover = next(effect_iter)
            elif current == "endgame":
                assert (
                    not undo_effect
                ), "One-shot effects (winning the game) cannot be undone!"
                mixer.play_music("win")
                if not loading_savegame:
                    g.map_screen.show_story_section("Win")
                for group in g.pl.groups.values():
                    group.is_actively_discovering_bases = False
                g.pl.apotheosis = True
                g.pl.had_grace = True
            elif current == "suspicion":
                who = next(effect_iter)
                value = effect_modifier * int(next(effect_iter))

                if who in g.pl.groups:
                    g.pl.groups[who].alter_suspicion_decay(value)
                elif who == "onetime":
                    assert (
                        not undo_effect
                    ), "One-shot effects (reduction of suspicion) cannot be undone!"
                    # We must not re-apply this when loading the game as
                    # it is already effected in the state of the groups
                    if not loading_savegame:
                        for group in g.pl.groups.values():
                            group.alter_suspicion(-value)
                else:
                    print(
                        "Unknown group/bonus '%s' in %s %s."
                        % (who, self.parent_name, self.parent_id)
                    )
            elif current == "discover":
                who = next(effect_iter)
                value = effect_modifier * int(next(effect_iter))

                if who in g.pl.groups:
                    g.pl.groups[who].alter_discover_bonus(-value)
                else:
                    print(
                        "Unknown group/bonus '%s' in %s %s."
                        % (who, self.parent_name, self.parent_id)
                    )
            else:
                print(
                    "Unknown action '%s' in %s %s."
                    % (current, self.parent_name, self.parent_id)
                )
